----------------------------------------------------------------------------
-- @Author: ViperGTS96-----------------------------------------------------
----------------------------------------------------------------------------
--------------------"The simplest design is the best design." --------------
----------------------------------------------------------------------------
--This design of handling wheel particle systems is reserved by ViperGTS96--
----------------------------------------------------------------------------

snowFX = {};
snowFX.dir = ""	
local modDescFile = loadXMLFile("modDesc", g_currentModDirectory.. "modDesc.xml");
snowFX.title = getXMLString(modDescFile, "modDesc.title.en");
snowFX.author = getXMLString(modDescFile, "modDesc.author");
snowFX.version = getXMLString(modDescFile, "modDesc.version");
snowFX.printVehiclesToLog = false; --(debugging)
snowFX.maxWheelLimit = 10;
snowFX.psSpeedRatio = 15000; --Divedend
snowFX.psEmitRate = 0.4301; --Base value
delete(modDescFile);

function snowFX:setDir(dir) snowFX.dir = dir; end;

function snowFX.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(Washable, specializations)
		and SpecializationUtil.hasSpecialization(AnimatedVehicle, specializations);
end;

function snowFX.registerEventListeners(vehicleType)
	for _, functionName in pairs( { "onPostLoad", "onUpdate", "onLeave" } ) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, snowFX);
	end;
end;

function snowFX:onPostLoad(savegame)

	if snowFX.dir == "" then return; end;
	if self.spec_wheels == nil then return; end;
	local wheels = self.spec_wheels.wheels;
	local crawlers = self.spec_crawlers;
	if wheels ~= nil and #wheels > 0 then
		self.snow_FX = {};
		self.snow_FX.tooManyWheelsForsnow = false;
		if crawlers ~= nil and #crawlers.crawlers > 0 then
			crawlers = crawlers.crawlers;
			self.snow_FX.crawlers = #crawlers;
		end;
		if #wheels > snowFX.maxWheelLimit then
			self.snow_FX.tooManyWheelsForsnow = true;
			return;
		else
			self.getDirtMultiplier = Utils.overwrittenFunction(Washable.getDirtMultiplier, snowFX.getDirtMultiplier);
			for index, wheel in pairs(wheels) do		
				local wheelLinkNode = wheel.driveNode;
				if wheelLinkNode == nil then wheelLinkNode = wheel.repr; end;
				wheel.snowFX_isCrawler = snowFX:getIsCrawler(self, wheel, crawlers);
				local hasIssue, fix = snowFX:getHasIssue(self, index, wheel);
				if not hasIssue then
					local snow_psData = {};
					local scaleFactor = 1.0;
					wheel.snow_psF = {};
					wheel.snow_psR = {};
					wheel.snowFXisActive = false;
					wheel.snowFXMainNode = createTransformGroup("snowNode");
					link(wheel.node, wheel.snowFXMainNode);
					local rDX1,rDY1,rDZ1 = localToWorld(self.components[1].node, 0,0,1);
					local rDX2,rDY2,rDZ2 = localToWorld(self.components[1].node, 0,1,0);
					setWorldTranslation(wheel.snowFXMainNode, unpack({localToWorld(wheelLinkNode,0,-wheel.radius,0)}));
					setDirection(wheel.snowFXMainNode,rDX1,rDY1,rDZ1,rDX2,rDY2,rDZ2);
					if fix == 1 then local bx,by,bz = getTranslation(wheels[index-1].snowFXMainNode);
					setTranslation(wheel.snowFXMainNode, -bx,by,bz);	end;
					snow_psData.psFile = "res/snowPS.i3d";
					snow_psData.posX = 0;
					snow_psData.posY = 0;
					snow_psData.posZ = -wheel.radius;	
					snow_psData.worldSpace = false;
					ParticleUtil.loadParticleSystemFromData(snow_psData, wheel.snow_psF, nil, false, nil, snowFX.dir, wheel.snowFXMainNode);
					snow_psData.posZ = wheel.radius;
					if wheel.snowFX_isCrawler then snow_psData.posZ = crawlers[1].trackWidth*3 end;
					ParticleUtil.loadParticleSystemFromData(snow_psData, wheel.snow_psR, nil, false, nil, snowFX.dir, wheel.snowFXMainNode);
					local rX,rY,rZ = getRotation(wheel.snow_psR.shape);
					setRotation(wheel.snow_psR.shape, rX,rY-math.rad(180),rZ);
					setScale(wheel.snow_psF.shape, wheel.width * scaleFactor, wheel.radius * scaleFactor, wheel.radius * scaleFactor);
					setScale(wheel.snow_psR.shape, wheel.width * scaleFactor, wheel.radius * scaleFactor, wheel.radius * scaleFactor);
					wheel.snow_psF.resetPos = {getTranslation(wheel.snow_psF.shape)}
					wheel.snow_psR.resetPos = {getTranslation(wheel.snow_psR.shape)}
					local ir = 1;
					local il = 1;
					if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
						for _,additionalWheel in pairs (wheel.additionalWheels) do
							additionalWheel.snowFX_isCrawler = false;
							if additionalWheel.tireType == WheelsUtil.getTireType("crawler") then
								additionalWheel.snowFX_isCrawler = true;
							end;
							local offset = 0.11*wheel.radius;
							if wheel.xOffset ~= nil and wheel.xOffset ~= 0 then offset = wheel.xOffset; end;
							if additionalWheel.isLeft then
								snow_psData.posX = ((wheel.width*il) + offset);
								il = il + 1;
							else
								snow_psData.posX = -((wheel.width*il) + offset);
								ir = ir + 1;
							end;
							additionalWheel.snow_psF = {};
							additionalWheel.snow_psR = {};
							snow_psData.posZ = -wheel.radius;
							ParticleUtil.loadParticleSystemFromData(snow_psData, additionalWheel.snow_psF, nil, false, nil, snowFX.dir, wheel.snowFXMainNode);
							snow_psData.posZ = wheel.radius;
							if additionalWheel.snowFX_isCrawler then snow_psData.posZ = crawlers[1].trackWidth*3 end;
							ParticleUtil.loadParticleSystemFromData(snow_psData, additionalWheel.snow_psR, nil, false, nil, snowFX.dir, wheel.snowFXMainNode);
							local rX,rY,rZ = getRotation(additionalWheel.snow_psR.shape);
							setRotation(additionalWheel.snow_psR.shape, rX,rY-math.rad(180),rZ);
							setScale(additionalWheel.snow_psF.shape, additionalWheel.width * scaleFactor, additionalWheel.radius * scaleFactor, additionalWheel.radius * scaleFactor);
							setScale(additionalWheel.snow_psR.shape, additionalWheel.width * scaleFactor, additionalWheel.radius * scaleFactor, additionalWheel.radius * scaleFactor);
							additionalWheel.snow_psF.resetPos = {getTranslation(additionalWheel.snow_psF.shape)}
							additionalWheel.snow_psR.resetPos = {getTranslation(additionalWheel.snow_psR.shape)}
						end;
					end;
				end;
			end;
		end;
		if snowFX.printVehiclesToLog and self.snow_FX ~= nil then print("Load mod: '"..snowFX.title.."' successfully applied to "..FSBaseMission:getVehicleName(self)); end;
	end;
end;

function snowFX:getIsCrawler(self, wheel, crawlers)	
	if wheel.wheelTire == nil then
		if self.snow_FX.crawlers ~= nil then
			if wheel.tireType == WheelsUtil.getTireType("crawler") then 
				return true; 
			else
				for _,crawler in pairs(crawlers) do
					if crawler.wheel == wheel then 
						return true; 
					end;
				end;
			end;
		end;
	end;
	return false;
end;

function snowFX:getHasIssue(self, index, wheel)
	if not wheel.hasParticles and wheel.wheelTire == nil then 
		return true, 0; 
	end;
	---------------
	if self.spec_plow ~= nil or self.spec_cultivator ~= nil or self.spec_roller then 
		return true, 0; 
	end;
	---------------
	local namesToFind = {"jcb fastrac 4220"}; --wheel[4] has local offset from rotation
	local vName = FSBaseMission:getVehicleName(self);
	vName = string.lower(vName);
	for _, nameToFind in pairs(namesToFind) do
		if string.find(vName,nameToFind) and index > 3 then 
			return false, 1;
		end;
	end;
	---------------
	--local x,y,z = getRotation(wheel.driveNode);
	--if x+y+z ~= 0 then return true, 0; end;
	---------------
	local storeCat = g_storeManager:getItemByXMLFilename(self.configFileName);
	if storeCat ~= nil and storeCat.categoryName ~= nil then
		storeCat = storeCat.categoryName;
		for _, type in pairs({ "PLANTERS", "SEEDERS", "WEEDERS", "TEDDERS",
		"WINDROWERS", "PLOWS", "SUBSOILERS", "CULTIVATORS", "DISCHARROWS", "POWERHARROWS" }) do
			if type == storeCat then 
				return true, 0;				
			end;
		end;
	end;
	---------------
	return false, 0;
end;

function snowFX:getIsTireOnSnow(wheel)
	if g_currentMission.terrainRootNode ~= nil and g_currentMission.terrainDetailId ~= nil then
		local wheelNode = wheel.wheelTire;
		if wheelNode == nil then wheelNode = wheel.driveNode; end; --crawler
		if wheelNode == nil then wheelNode = wheel.repr; end;
		local x1,y1,z1 = localToWorld(wheelNode, -wheel.width/2,-wheel.radius,0);
		local x2,y2,z2 = localToWorld(wheelNode, wheel.width/2,-wheel.radius,0);
		y1 = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x1,y1,z1);
		y2 = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x2,y2,z2);
		local foundFillType = DensityMapHeightUtil.getFillTypeAtLine(x1,y1,z1, x2,y2,z2, 1.0);
		if foundFillType == FillType.SNOW then
			return true;
		end;
	end;
	return false;
end;

function snowFX:stopPS(wheel)
	if wheel.snow_psF ~= nil then ParticleUtil.setEmittingState(wheel.snow_psF, false, false, false); end;
	if wheel.snow_psR ~= nil then ParticleUtil.setEmittingState(wheel.snow_psR, false, false, false); end;
	wheel.snowFXisActive = false;
end;

function snowFX:stopEveryPS(self)
	if self.snow_FX ~= nil then
		local wheels = self.spec_wheels.wheels;	
		if wheels ~= nil and #wheels > 0 and not self.snow_FX.tooManyWheelsForsnow then
			for w=1, #wheels do
				local wheel = wheels[w];
				if wheel.snowFXisActive ~= nil then
					snowFX:stopPS(wheel);			
					if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
						for _,additionalWheel in pairs (wheel.additionalWheels) do
							snowFX:stopPS(additionalWheel);
						end;
					end;
				end;
			end;
		end;
	end;
end;

function snowFX:correctPSLocation(wheel)	
	local _,y,_ = getTranslation(wheel.snow_psF.shape);
	local _,y2,_ = getTranslation(wheel.snow_psR.shape);
	setTranslation(wheel.snow_psF.shape, wheel.snow_psF.resetPos[1], y, wheel.snow_psF.resetPos[3]);
	setTranslation(wheel.snow_psR.shape, wheel.snow_psR.resetPos[1], y2, wheel.snow_psR.resetPos[3]);
end;

function snowFX:setPSHeight(wheel)
	local x,y,z = getWorldTranslation(wheel.snowFXMainNode);
	local gY = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x,y,z);
	local xT,_,zT = getWorldTranslation(wheel.snow_psF.shape);
	local xT2,_,zT2 = getWorldTranslation(wheel.snow_psR.shape);
	setWorldTranslation(wheel.snow_psF.shape, xT,gY,zT);
	setWorldTranslation(wheel.snow_psR.shape, xT2,gY,zT2);
	snowFX:correctPSLocation(wheel);
	if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
		for _, aWheel in pairs(wheel.additionalWheels) do
			local xx,_,zz = getWorldTranslation(aWheel.snow_psF.shape);	
			local xx2,_,zz2 = getWorldTranslation(aWheel.snow_psR.shape);
			local xx3,yy3,zz3 = localToWorld(aWheel.snow_psF.shape,0,0,aWheel.radius/2)
			gY = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, xx3,yy3,zz3);
			setWorldTranslation(aWheel.snow_psF.shape, xx,gY,zz);
			setWorldTranslation(aWheel.snow_psR.shape, xx2,gY,zz2);	
			snowFX:correctPSLocation(aWheel);	
		end;
	end;
end;

function snowFX:updateWheelRotation(self, wheel)
	if self.snow_FX.crawlers == nil or wheel.wheelTire ~= nil then
		setRotation(wheel.snowFXMainNode,0,wheel.steeringAngle,0);
	end;
end;

function snowFX:updatePS(self, dt)
	local speedfactor = snowFX.psSpeedRatio;
	local enteredSpec = self.spec_enterable;
	local attachableSpec = self.spec_attachable;
	local isEntered = false;
	local isAttachment = false;
	if enteredSpec ~= nil then isEntered = Utils.getNoNil(enteredSpec.isEntered, false); end;
	if not isEntered then
		if attachableSpec ~= nil and attachableSpec.attacherVehicle ~= nil and attachableSpec.attacherVehicle.spec_enterable then
			isEntered = Utils.getNoNil(attachableSpec.attacherVehicle.spec_enterable.isEntered, false);
			isAttachment = true;
		end;
	end;
	if isEntered and not self.tooManyWheelsForsnow then		
			local wheels = self.spec_wheels.wheels;
			for w=1, #wheels do
				local wheel = wheels[w];
				if wheel.snowFXisActive ~= nil then
					local tspeed = 0;
					if isAttachment then tspeed = self:getLastSpeed(true);
					elseif wheel.netInfo.xDriveSpeed ~= nil then 
						tspeed = MathUtil.rpmToMps(wheel.netInfo.xDriveSpeed/(2*math.pi)*60,wheel.radius)*3600;
						if tspeed < 0 then tspeed = tspeed*-1; end;
					end;
					if not wheel.snowFXisActive then   --[[Activate]]
						if tspeed > 4.0 and snowFX:getIsTireOnSnow(wheel) then							
							if wheel.contact ~= Wheels.WHEEL_NO_CONTACT then wheel.snowFXisActive = true; end;
						end;
					else
						snowFX:setPSHeight(wheel);
						----------------------------
						if tspeed <= 4.0 then
							snowFX:stopPS(wheel);
						----------------------------
						elseif self.movingDirection < 0 then
						----------------------------Update Reverse--------------------------------------------------
							ParticleUtil.setEmittingState(wheel.snow_psF, false);
							ParticleUtil.setEmittingState(wheel.snow_psR, true);
							ParticleUtil.setParticleSystemSpeed(wheel.snow_psR, tspeed/(speedfactor/(wheel.radius*2)));	
						elseif self.movingDirection > 0 then
						----------------------------Update Forwards--------------------------------------------------
							ParticleUtil.setEmittingState(wheel.snow_psR, false);				
							ParticleUtil.setEmittingState(wheel.snow_psF, true);
							ParticleUtil.setParticleSystemSpeed(wheel.snow_psF, tspeed/(speedfactor/(wheel.radius*2)));						
						end;
					end;
					if not snowFX:getIsTireOnSnow(wheel) then
						snowFX:stopPS(wheel);
						snowFX:setPSHeight(wheel);
					end;
					------------------------------------------------------------------
					if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
						for _,additionalWheel in pairs (wheel.additionalWheels) do
							if not additionalWheel.snowFXisActive then
								if tspeed > 4.0 and snowFX:getIsTireOnSnow(additionalWheel) then
									additionalWheel.snowFXisActive = true;
								end;
							else
								----------------------------
								if tspeed <= 4.0 then
									snowFX:stopPS(additionalWheel);
								----------------------------
								elseif self.movingDirection < 0 then
								----------------------------Update Reverse--------------------------------------------------
									ParticleUtil.setEmittingState(additionalWheel.snow_psF, false);
									ParticleUtil.setEmittingState(additionalWheel.snow_psR, true);
									ParticleUtil.setParticleSystemSpeed(additionalWheel.snow_psR, tspeed/(speedfactor/(additionalWheel.radius*2)));	
								elseif self.movingDirection > 0 then
								----------------------------Update Forwards--------------------------------------------------
									ParticleUtil.setEmittingState(additionalWheel.snow_psR, false);
									ParticleUtil.setEmittingState(additionalWheel.snow_psF, true);
									ParticleUtil.setParticleSystemSpeed(additionalWheel.snow_psF, tspeed/(speedfactor/(additionalWheel.radius*2)));
								end;
							end;
							if not snowFX:getIsTireOnSnow(additionalWheel) then 
								snowFX:stopPS(additionalWheel);
							end;
						end;
					end;
					----------------------------------------------------------------------
					snowFX:updateWheelRotation(self, wheel);
				end;
			end;
	elseif not self.tooManyWheelsForsnow then
		snowFX:stopEveryPS(self);
	end;

end;

function snowFX:onUpdate(dt)
	if self.spec_wheels ~= nil then
		if self.spec_wheels.wheels ~= nil and #self.spec_wheels.wheels > 0 then
			snowFX:updatePS(self, dt);
		end;
	end;
end;

function snowFX:onLeave()
	snowFX:stopEveryPS(self);
end;

function snowFX:getDirtMultiplier()
    local spec = self.spec_washable
    local multiplier = 1
    if self:getLastSpeed() < 1 then
        multiplier = 0
    end
    if self:getIsOnField() then
        multiplier = multiplier * spec.fieldMultiplier
        local wetness = g_currentMission.environment.weather:getGroundWetness()
        if wetness > 0 then
            multiplier = multiplier * (1 + wetness)
        end
		local isInsnow = false;
		if self.spec_wheels ~= nil and self.spec_wheels.wheels ~= nil then
			for _, wheel in pairs(self.spec_wheels.wheels) do
				if snowFX:getIsTireOnSnow(wheel) then
					isInsnow = true;
					break;
				end;
			end;
		end;
		if isInsnow then multiplier = multiplier * 10; end;
    end	
    return multiplier
end;

function snowFX:disable()
	snowFX.disabled=1;
end;

if snowFX.disabled == nil then
	print("Load mod: "..snowFX.title.."'s SnowFX : v"..snowFX.version.." by "..snowFX.author);
end;